import argparse
from typing import Optional


class ObjectWithArgparsedArgs:
    """Class of objects with arguments defined by `argparse`.

    If `args` is not given, all arguments are set using `argparse`.

    Args:
        args (argparse.Namespace, optional): Arguments.
    """

    def __init__(self, args: Optional[argparse.Namespace] = None, **kwargs):
        if args is None:
            parser = argparse.ArgumentParser(
                formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                allow_abbrev=False)
            self._default_args = self._define_args_defaults()
            self._add_args(parser)
            args = parser.parse_args()
        self._check_args(args)
        self._args = args

    @classmethod
    def _define_args_defaults(cls) -> dict:
        """Defines defaults of command-line arguments for this class.

        A subclass must override this method only if 1) the subclass introduces
        new arguments for which defaults should be set, or 2) the defaults
        defined here have to be updated.

        All subclasses that override this method *must* call the super method
        at the beginning of the overriding implementation. The returned
        dictionary can be updated before being returned.

        Returns:
            defaults (dict): a dictionary of argument names and default values.
        """
        return {}

    def _add_args(self, parser: argparse.ArgumentParser) -> None:
        """Adds class-specific arguments to the argument parser.

        Subclasses overriding this methods *must* call the super method.

        Args:
            parser (argparse.ArgumentParser): Argument parser automatically
                created when initializing this object.
        """
        pass

    @classmethod
    def _check_args(cls, args: argparse.Namespace) -> None:
        """Checks arguments relevant to this class.

        If a subclass overrides this method, it should check its own arguments,
        and it *must* call the super's implementation of this method.

        Args:
            args (argparse.Namespace)
        """
        pass

    @property
    def args(self) -> argparse.Namespace:
        """Arguments (configuration) of this object."""
        return self._args



import argparse
import os
import pickle
import warnings
from typing import Type, Optional

import torch

from boilr.experiments import BaseExperimentManager
from boilr.utils import viz
from boilr.utils.meta import ObjectWithArgparsedArgs
from boilr.utils.utils import get_date_str


class BaseOfflineEvaluator(ObjectWithArgparsedArgs):
    """Boilerplate code to run evaluation routines on a trained model.

    Initialize with the experiment class used for training (the class, not the
    instance).

    A subclass must define:
    - the _parse_args() method, which is called during init, and must include a
      call to add_required_args() with the argparser as argument.
    - the run() method, defining the whole evaluation procedure.

    The evaluator can then be called directly because it implements __call__().

    The following data attributes are defined, to be used by subclasses:
    - self._experiment
    - self._img_folder

    Args:
        args (argparse.Namespace, optional): Unused.
        experiment_class (class): Class object of (a subclass of)
            `BaseExperimentManager`.
    """

    def __init__(self,
                 args: Optional[argparse.Namespace] = None,
                 experiment_class: Type[BaseExperimentManager] = None):
        del args
        if experiment_class is None:
            raise ValueError("Argument `experiment_class` cannot be None")
        super(BaseOfflineEvaluator, self).__init__()

        args = self.args  # these are the evaluation args
        use_cuda = not args.no_cuda and torch.cuda.is_available()
        device = torch.device("cuda" if use_cuda else "cpu")
        date_str = get_date_str()
        print('device: {}, start time: {}'.format(device, date_str))

        # Get path to load model
        checkpoint_folder = os.path.join('output', args.load, 'checkpoints')

        # Add date string and create folder on evaluation_results
        self._result_folder = os.path.join('output', args.load,
                                           'evaluation_output',
                                           'eval_' + date_str)
        self._img_folder = os.path.join(self._result_folder, 'imgs')
        os.makedirs(self._result_folder)
        os.makedirs(self._img_folder)

        # Set img folder for viz module
        viz.img_folder = self._img_folder

        # Load config (this contains the *original* experiment args)
        config_path = os.path.join(checkpoint_folder, 'config.pkl')
        with open(config_path, 'rb') as file:
            experiment_args = pickle.load(file)

        # Modify experiment config for testing
        if args.test_batch_size is not None:
            experiment_args.test_batch_size = args.test_batch_size
        experiment_args.dry_run = False

        # Create and set up experiment manager
        experiment = experiment_class(args=experiment_args)
        experiment.setup(device=device, checkpoint_folder=checkpoint_folder)

        self._experiment = experiment

    @classmethod
    def _define_args_defaults(cls) -> dict:
        defaults = super(BaseOfflineEvaluator, cls)._define_args_defaults()
        defaults.update(
            default_run="",
            test_batch_size=None,
        )
        return defaults

    def _add_args(self, parser: argparse.ArgumentParser) -> None:
        super(BaseOfflineEvaluator, self)._add_args(parser)

        parser.add_argument('--load',
                            type=str,
                            metavar='NAME',
                            default=self._default_args['default_run'],
                            help="name of the run to be loaded")
        parser.add_argument('--load-step',
                            type=int,
                            dest='load_step',
                            metavar='N',
                            help='step of checkpoint to be loaded (default:'
                            ' last available)')
        parser.add_argument('--test-batch-size',
                            type=int,
                            default=self._default_args['test_batch_size'],
                            dest='test_batch_size',
                            metavar='N',
                            help='test batch size')
        parser.add_argument('--nocuda',
                            action='store_true',
                            dest='no_cuda',
                            help='do not use cuda')

    @classmethod
    def _check_args(cls, args: argparse.Namespace) -> None:
        super(BaseOfflineEvaluator, cls)._check_args(args)
        if args.load_step is not None:
            warnings.warn(
                "Loading weights from specific training step is not supported "
                "for now. The model will be loaded from the last checkpoint.")

    def run(self):
        """Runs the evaluator."""
        raise NotImplementedError

    def __call__(self, *args, **kwargs):
        """Runs the evaluator."""
        self.run()



import math
from collections import OrderedDict

import torch

from boilr.options import get_option

__sentinel = object()


def print_num_params(model, max_depth=__sentinel):
    """Prints overview of model architecture with number of parameters.

    Optionally, it groups together parameters below a certain depth in the
    module tree. The depth defaults to packagewide options.

    Args:
        model (torch.nn.Module)
        max_depth (int, optional)
    """

    if max_depth is __sentinel:
        max_depth = get_option('model_print_depth')
    assert max_depth is None or isinstance(max_depth, int)

    sep = '.'  # string separator in parameter name
    print("\n--- Trainable parameters:")
    num_params_tot = 0
    num_params_dict = OrderedDict()

    for name, param in model.named_parameters():
        if not param.requires_grad:
            continue

        num_params = param.numel()

        if max_depth is not None:
            split = name.split(sep)
            prefix = sep.join(split[:max_depth])
        else:
            prefix = name
        if prefix not in num_params_dict:
            num_params_dict[prefix] = 0
        num_params_dict[prefix] += num_params
        num_params_tot += num_params
    for n, n_par in num_params_dict.items():
        print("{:8d}  {}".format(n_par, n))
    print("  - Total trainable parameters:", num_params_tot)
    print("---------\n")


def grad_global_norm(parameters, norm_type=2):
    """Compute global norm of the gradients of an iterable of parameters.

    The norm is computed over all gradients together, as if they were
    concatenated into a single vector.

    Args:
        parameters (Iterable[Tensor] or Tensor): an iterable of Tensors or a
            single Tensor
        norm_type (float or int, optional): type of the used p-norm. Can be
            ``'inf'`` for infinity norm.

    Returns:
        Global norm of the parameters' gradients (viewed as a single vector).
    """
    if isinstance(parameters, torch.Tensor):
        parameters = [parameters]
    grads = [p.grad for p in parameters if p.grad is not None]
    return global_norm(grads, norm_type=norm_type)


def global_norm(parameters, norm_type=2):
    """Compute global norm of an iterable of parameters.

    The norm is computed over all tensors together, as if they were
    concatenated into a single vector. This code is based on
    torch.nn.utils.clip_grad_norm_().

    Args:
        parameters (Iterable[Tensor] or Tensor): an iterable of Tensors or a
            single Tensor
        norm_type (float or int, optional): type of the used p-norm. Can be
            ``'inf'`` for infinity norm.

    Returns:
        Total norm of the parameters (viewed as a single vector).
    """
    if isinstance(parameters, torch.Tensor):
        parameters = [parameters]
    norm_type = float(norm_type)
    if norm_type == math.inf:
        total_norm = max(p.data.abs().max() for p in parameters)
    else:
        total_norm = 0.
        for p in parameters:
            param_norm = p.data.norm(norm_type)
            total_norm += param_norm.item()**norm_type
        total_norm = total_norm**(1. / norm_type)
    return total_norm


def get_module_device(module):
    """Returns the module's device.

    This is a simple trick, it is probably not robust.

    Args:
        module (torch.nn.Module)

    Returns:
        device (torch.device)
    """
    return next(module.parameters()).device


def is_conv(module):
    """Returns whether the module is a convolutional layer."""
    return isinstance(module, torch.nn.modules.conv._ConvNd)


def is_linear(module):
    """Returns whether the module is a linear layer."""
    return isinstance(module, torch.nn.Linear)


def named_leaf_modules(module):
    """Yields (name, module) pairs that are leaves in the module tree."""

    # Should work under common naming assumptions, but it's not guaranteed
    last_name = ''
    for name, l in reversed(list(module.named_modules())):
        if name not in last_name:
            last_name = name
            yield name, l



import os
from collections import OrderedDict

import numpy as np
import torch
from matplotlib import pyplot as plt
from torchvision.utils import make_grid
from torchvision.utils import save_image

from boilr.nn.utils import named_leaf_modules

img_folder = None


def _unique_filename(fname, extension):

    def exists(fname):
        return os.path.exists(fname + '.' + extension)

    if not exists(fname):
        return fname
    for i in range(10000):
        t = fname + '_' + str(i)
        if not exists(t):
            return t
    raise RuntimeError("too many files ({})".format(i))


def plot_imgs(imgs, name=None, extension='png', colorbar=True, overwrite=False):
    """Saves collection of images.

    Plots collection of 1-channel images (as 3D tensor, or 4D tensor with size
    1 on the 2nd dimension) and saves it as png. If any image extends beyond
    [0, 1], all are normalized such that the minimum and maximum are 0 and 1.

    If overwrite is False, it automatically appends an integer to the filename
    to make it unique.

    Args:
        imgs (torch.Tensor):
        name (str, optional):
        extension (str, optional):
        colorbar (bool, optional):
        overwrite (bool, optional):
    """
    if imgs.dim() == 4 and imgs.size(1) == 1:
        imgs = imgs.squeeze(1)
    if imgs.dim() != 3:
        msg = ("input tensor must be 3D, or 4D with size 1 on the 2nd "
               "dimension, but has shape {}".format(imgs.shape))
        raise RuntimeError(msg)
    if img_folder is None:
        raise RuntimeError("Image folder not set")
    fname = name.replace(' ', '_')
    fname = os.path.join(img_folder, fname)
    if not overwrite:
        fname = _unique_filename(fname, extension)
    fname = fname + '.' + extension
    imgs = imgs.detach().cpu().unsqueeze(1)  # (N, 1, H, W)
    n_imgs = imgs.size(0)
    _, c = balanced_approx_factorization(n_imgs)  # grid arrangement

    # Get minimum and maximum
    low = imgs.min().item()
    high = imgs.max().item()

    # Normalize if images extend beyond the range [0, 1]
    normalize = low < 0. or high > 1.
    if normalize:
        imgs = (imgs - low) / (high - low)

    # Compute pad value, either 0 or 1
    pad_value = img_grid_pad_value(imgs)

    # Images are now in [0, 1], arrange them into a grid as they are
    # Grid has shape (3, grid_h, grid_w) and has values in [0, 1]
    grid = make_grid(imgs, nrow=c, pad_value=pad_value, normalize=False)

    if colorbar:
        # make_grid made it RGB, now take one channel only
        grid = grid[0]

        # Rescale images to original interval (now including the padding)
        if normalize:
            grid = grid * (high - low) + low

        # Save grid of images with colorbar
        vmin = low if normalize else 0.
        vmax = high if normalize else 1.
        plt.imshow(grid, vmin=vmin, vmax=vmax, cmap='gray')
        plt.colorbar()
        plt.title(name)
        plt.savefig(fname)
        plt.close()

    else:
        from PIL import Image

        # Add 0.5 after unnormalizing to [0, 255] to round to nearest integer
        grid = grid.mul(255).add_(0.5).clamp_(0, 255).to('cpu', torch.uint8)
        grid = grid.permute(1, 2, 0).numpy()

        # Save image with PIL
        im = Image.fromarray(grid)
        im.save(fname, format=None)


def img_grid_pad_value(imgs, thresh=.2):
    """Returns padding value (black or white) for a grid of images.

    Hack to visualize boundaries between images with torchvision's
    save_image(). If the median border value of all images is below the
    threshold, use white, otherwise black (which is the default).

    Args:
        imgs (torch.Tensor): A 4d tensor
        thresh (float, optional): Threshold in (0, 1).

    Returns:
        pad_value (float): The padding value
    """

    assert imgs.dim() == 4
    imgs = imgs.clamp(min=0., max=1.)
    assert 0. < thresh < 1.

    imgs = imgs.mean(1)  # reduce to 1 channel
    h = imgs.size(1)
    w = imgs.size(2)
    borders = list()
    borders.append(imgs[:, 0].flatten())
    borders.append(imgs[:, h - 1].flatten())
    borders.append(imgs[:, 1:h - 1, 0].flatten())
    borders.append(imgs[:, 1:h - 1, w - 1].flatten())
    borders = torch.cat(borders)
    if torch.median(borders) < thresh:
        return 1.0
    return 0.0


def save_image_grid(images, filename, nrows=8):
    """Saves a grid of images.

    Wrapper around torchvision save_image. The grid lines are either black or
    white depending on the image colors.

    Args:
        images (Tensor): batch of images (N, channels, H, W)
        filename (str): name of output file
        nrows (int, optional): number of rows and columns. The number of images
            is the square of this.
    """
    pad = img_grid_pad_value(images)
    save_image(images, filename, nrow=nrows, pad_value=pad)


def save_image_grid_reconstructions(inputs, recons, filename):
    """Saves a grid of images with inputs and reconstructions.

    This is meant for autoencoder-like models. The number of images is inferred
    from the input sizes. Inputs and reconstructions must have the same shape.

    Args:
        inputs (Tensor): batch of input images
        recons (Tensor): batch of reconstructions
        filename (str): name of output file
    """
    assert inputs.shape == recons.shape
    n_img = inputs.shape[0]
    n = int(np.sqrt(2 * n_img))
    imgs = torch.stack([inputs.cpu(), recons.cpu()])
    imgs = imgs.permute(1, 0, 2, 3, 4)
    imgs = imgs.reshape(n**2, inputs.size(1), inputs.size(2), inputs.size(3))
    save_image_grid(imgs, filename, nrows=n)


def balanced_approx_factorization(x, ratio=1):
    """Approximately factorize an integer into integers.

    Useful for example to plot images in a grid.

    Args:
        x (int): Number to be approximately factorized
        ratio (float or int, optional): Ratio columns/rows

    Returns:
        dims (tuple): A tuple (rows, columns)
    """

    # We want c/r to be approx equal to ratio, and r*c to be approx equal to x
    # ==> r = x/c = x/(ratio*r)
    # ==> r = sqrt(x/ratio) and c = sqrt(x*ratio)
    assert type(x) == int or x.dtype == int
    c = int(np.ceil(np.sqrt(x * ratio)))
    r = int(np.ceil(x / c))
    return r, c


def balanced_factorization(x):
    """
    Util to plot images in a grid.

    :param x: number to be factorized
    """

    assert type(x) == int or x.dtype == int
    a = int(np.floor(np.sqrt(x)))
    while True:
        b = x / a
        if b.is_integer():
            return int(b), a
        a += 1


def clean_axes():
    """Cleans current axes."""
    plt.gca().tick_params(
        axis='both',  # changes apply to the x-axis
        which='both',  # both major and minor ticks
        left=False,  # ticks along the left edge are off
        right=False,
        bottom=False,
        top=False,
        labelleft=False,  # labels along the left edge are off
        labelbottom=False)


def _save_activation_hook(ord_dict, name):

    def hook(model, inp, ret):
        if isinstance(ret, tuple):
            try:
                ret = torch.cat(ret, dim=1)
            except TypeError:
                ret = ret[0]
            except RuntimeError as e:
                print("WARNING:", e)
                return
        ord_dict[name] = ret.detach()

    return hook


def set_up_saving_all_activations(model):
    """Registers forward hooks to save all activations in a model.

    It returns a dictionary that will be populated by the forward hooks at
    the next forward pass.

    Args:
        model (torch.nn.Module)

    Returns:
        activations (dict): A dict that will be populated with all activations
            at the next forward pass.
    """
    all_activations = OrderedDict()
    for module_name, module in named_leaf_modules(model):
        module.register_forward_hook(
            _save_activation_hook(all_activations, module_name))
    return all_activations


#########################
# Test

if __name__ == '__main__':

    ### Test plot imgs

    img_folder = ''

    low = -20.
    high = 30.
    imgs = torch.rand(4, 1, 8, 8) * (high - low) + low
    plot_imgs(imgs, name='test_colorbar_false', colorbar=False)
    plot_imgs(imgs, name='test_colorbar_true', colorbar=True)

    low = .7
    high = 2.
    imgs = torch.rand(4, 1, 8, 8) * (high - low) + low
    plot_imgs(imgs, name='test_colorbar_false', colorbar=False)
    plot_imgs(imgs, name='test_colorbar_true', colorbar=True)

    low = .5
    high = .9
    imgs = torch.rand(4, 1, 8, 8) * (high - low) + low
    plot_imgs(imgs, name='test_colorbar_false', colorbar=False)
    plot_imgs(imgs, name='test_colorbar_true', colorbar=True)

    ### Test img grid pad value
    img_grid_pad_value(torch.rand(6, 3, 32, 32), thresh=.3)



import torch
from torch import nn

from boilr.nn.utils import is_conv, is_linear
from boilr.utils import to_np

debug = False


def _get_data_dep_hook(init_scale):
    """Creates forward hook for data-dependent initialization.

    The hook computes output statistics of the layer, corrects weights and
    bias, and corrects the output accordingly in-place, so the forward pass
    can continue.

    Args:
        init_scale (float): Desired scale (standard deviation) of each
            layer's output at initialization.

    Returns:
        Forward hook for data-dependent initialization
    """

    def hook(module, inp, out):
        inp = inp[0]

        out_size = out.size()

        if is_conv(module):
            separation_dim = 1
        elif is_linear(module):
            separation_dim = -1
        dims = tuple([i for i in range(out.dim()) if i != separation_dim])
        mean = out.mean(dims, keepdim=True)
        var = out.var(dims, keepdim=True)

        if debug:
            print("Shapes:\n   input:  {}\n   output: {}\n   weight: {}".format(
                inp.size(), out_size, module.weight.size()))
            print("Dims to compute stats over:", dims)
            print("Input statistics:\n   mean: {}\n   var: {}".format(
                to_np(inp.mean(dims)), to_np(inp.var(dims))))
            print("Output statistics:\n   mean: {}\n   var: {}".format(
                to_np(out.mean(dims)), to_np(out.var(dims))))
            print("Weight statistics:   mean: {}   var: {}".format(
                to_np(module.weight.mean()), to_np(module.weight.var())))

        # Given channel y[i] we want to get
        #   y'[i] = (y[i]-mu[i]) * is/s[i]
        #         = (b[i]-mu[i]) * is/s[i] + sum_k (w[i, k] * is / s[i] * x[k])
        # where * is 2D convolution, k denotes input channels, mu[i] is the
        # sample mean of channel i, s[i] the sample variance, b[i] the current
        # bias, 'is' the initial scale, and w[i, k] the weight kernel for input
        # k and output i.
        # Therefore the correct bias and weights are:
        #   b'[i] = is * (b[i] - mu[i]) / s[i]
        #   w'[i, k] = w[i, k] * is / s[i]
        # And finally we can modify in place the output to get y'.

        scale = torch.sqrt(var + 1e-5)

        # Fix bias
        module.bias.data = ((module.bias.data - mean.flatten()) * init_scale /
                            scale.flatten())

        # Get correct dimension if transposed conv
        transp_conv = getattr(module, 'transposed', False)
        ch_out_dim = 1 if transp_conv else 0  # TODO handle groups

        # Fix weight
        size = tuple(-1 if i == ch_out_dim else 1 for i in range(out.dim()))
        weight_size = module.weight.size()
        module.weight.data *= init_scale / scale.view(size)
        assert module.weight.size() == weight_size

        # Fix output in-place so we can continue forward pass
        out.data -= mean
        out.data *= init_scale / scale

        assert out.size() == out_size

    return hook


def data_dependent_init(model, model_input_dict, init_scale=.1):
    """Performs data-dependent initialization on a model.

    Updates each layer's weights such that its outputs, computed on a batch
    of actual data, have mean 0 and the same standard deviation. See the code
    for more details.

    Args:
        model (torch.nn.Module):
        model_input_dict (dict): Dictionary of inputs to the model.
        init_scale (float, optional): Desired scale (standard deviation) of
            each layer's output at initialization. Default: 0.1.
    """

    hook_handles = []
    modules = filter(lambda m: is_conv(m) or is_linear(m), model.modules())
    for module in modules:
        # Init module parameters before forward pass
        nn.init.kaiming_normal_(module.weight.data)
        module.bias.data.zero_()

        # Forward hook: data-dependent initialization
        hook_handle = module.register_forward_hook(
            _get_data_dep_hook(init_scale))
        hook_handles.append(hook_handle)

    # Forward pass one minibatch
    model(**model_input_dict)  # dry-run

    # Remove forward hooks
    for hook_handle in hook_handles:
        hook_handle.remove()


#### Test #####

if __name__ == '__main__':
    debug = True

    # Test simple data-dependent init


    def do_test(x, layer):
        layer.bias.data.zero_()
        print("Output stats before:",
              layer(x).mean().item(),
              layer(x).var().item())
        handle = layer.register_forward_hook(_get_data_dep_hook(init_scale=0.5))
        y = layer(x)
        print("Output stats after:", y.mean().item(), y.var().item())
        handle.remove()

    # shape 64, 3, 5, 5
    x__ = (torch.rand(64, 3, 5, 5) - 0.2) * 20

    # Test Conv2d
    print("\n\n *** TEST Conv2d\n")
    do_test(x__, nn.Conv2d(3, 4, 3, padding=1))

    # Test ConvTranspose2d
    print("\n\n *** TEST ConvTranspose2d\n")
    do_test(x__, nn.ConvTranspose2d(3, 4, 3, padding=1))

    # Test Linear
    print("\n\n *** TEST Linear\n")
    x__ = x__.view(64, 25 * 3)  # flatten
    do_test(x__, nn.Linear(25 * 3, 8))



import re
from os.path import join

import setuptools

# Get version string by parsing _version.py
version_fname = join('boilr', '_version.py')
content = open(version_fname, "rt").read()
regex = r"^__version__ = ['\"]([^'\"]*)['\"]"
match = re.search(regex, content, re.M)
if match:
    version_string = match.group(1)
else:
    raise RuntimeError(
        "Unable to find version string in {}".format(version_fname))

# Get long description from README
with open("README.md", "r") as fh:
    long_description = fh.read()

# Define list of packages
packages = setuptools.find_packages()
packages = [p for p in packages if p.startswith('boilr')]

setuptools.setup(name="boilr",
                 version=version_string,
                 author="Andrea Dittadi",
                 author_email="andrea.dittadi@gmail.com",
                 description="Basic framework for training models with PyTorch",
                 long_description=long_description,
                 long_description_content_type="text/markdown",
                 url="https://github.com/addtt/boiler-pytorch",
                 packages=packages,
                 classifiers=[
                     "Programming Language :: Python :: 3",
                     "License :: OSI Approved :: MIT License",
                     "Operating System :: OS Independent",
                 ],
                 python_requires='>=3.6',
                 install_requires=[
                     'numpy>=1.18',
                     'torch>=1.4',
                     'torchvision>=0.5',
                     'matplotlib>=3.2',
                     'tqdm',
                     'pillow>=7.1',
                 ])



from torch.utils.data import DataLoader


class BaseDatasetManager:
    """Wrapper for DataLoaders.

    This is meant to be subclassed for specific experiments and datasets.
    The method _make_datasets() must be implemented. A basic implementation of
    the method _make_dataloaders() is provided in this class, but can be
    overridden for custom behavior.

    Args:
        cfg (argparse.Namespace): Configuration
        cuda (bool): Whether the device in use is cuda
    """

    def __init__(self, cfg, cuda, **kwargs):

        # Define training and test set
        tr_set, ts_set = self._make_datasets(cfg, **kwargs)

        # Dataloaders
        self._train, self._test = self._make_dataloaders(
            tr_set, ts_set, cfg, cuda, **kwargs)

        self._data_shape = self._train.dataset[0][0].size()
        self._img_size = self._data_shape[1:]
        self._color_ch = self._data_shape[0]

    @classmethod
    def _make_datasets(cls, cfg, **kwargs):
        """Returns training and test sets as PyTorch Datasets.

        Args:
            cfg (argparse.Namespace): Configuration
        """
        raise NotImplementedError

    @classmethod
    def _make_dataloaders(cls, train, test, cfg, cuda, **kwargs):
        """Returns training and test data loaders.

        Default data loaders provided here. Override for custom data loaders.

        Args:
            train (Dataset): Training set
            test (Dataset): Test set
            cfg (argparse.Namespace): Configuration
            cuda (bool): Whether the device in use is cuda

        Returns:
            (tuple): tuple containing:
                - dl_train (DataLoader): training set data loader
                - dl_test (DataLoader): test set data loader
        """

        # Default arguments for dataloaders
        nw = getattr(kwargs, 'num_workers', 0)
        pm = getattr(kwargs, 'pin_memory', False)
        dl_kwargs = {'num_workers': nw, 'pin_memory': pm} if cuda else {}

        dl_train = DataLoader(train,
                              batch_size=cfg.batch_size,
                              shuffle=True,
                              drop_last=True,
                              **dl_kwargs)
        dl_test = DataLoader(test,
                             batch_size=cfg.test_batch_size,
                             shuffle=False,
                             **dl_kwargs)
        return dl_train, dl_test

    @property
    def train(self):
        """DataLoader for training set"""
        return self._train

    @property
    def test(self):
        """DataLoader for test set"""
        return self._test

    @property
    def data_shape(self):
        """Shape of each datapoint (image): (channels, height, width)."""
        return self._data_shape

    @property
    def img_size(self):
        """Spatial shape of each datapoint (image): height, width."""
        return self._img_size

    @property
    def color_ch(self):
        """Number of color channels of each datapoint (image)."""
        return self._color_ch
